# Gridlock - a collection of grid-based strategy games
# Copyright (c) 2009 Brian Nenninger (brian@dozingcatsoftware.com)
# Gridlock comes with ABSOLUTELY NO WARRANTY, and is released under version 3 of the GNU GPL.

import GridUtils
from Game import GameState

class ZoneshGameState(GameState):
    def reset(self):
        GameState.reset(self)
        self.extra_state = {'goal_positions': [None, None]}
        self.extra_state['goal_positions'][1] = (0,) * self.grid.ndim
        self.extra_state['goal_positions'][0] = tuple(d-1 for d in self.grid.shape)
        # if positions not provided, fill by zones
        if 'positions' not in self.game_params:
            for pos in self.itercoords():
                self.grid[pos] = self.zone_for_position(pos)
        
    def has_scores(self):
        return False

    def zone_for_position(self, pos):
        # compute "manhattan distance" from corners and compare to zone_size
        dist1 = sum(pos)
        if dist1<self.game_params['zone_size']: return 1
        dist2 = sum(self.grid.shape) - self.grid.ndim - dist1
        if dist2<self.game_params['zone_size']: return 2
        return 0
        
    # HACK: when serialized, goal_positions gets turned into lists, but inner values need to be tuples
    def finish_deserialization(self, obj):
        for i,pos in enumerate(self.extra_state['goal_positions']):
            self.extra_state['goal_positions'][i] = tuple(pos)

    def is_game_over(self):
        return self.grid[self.extra_state['goal_positions'][self.previous_player()-1]]==self.previous_player() \
               or self.current_player not in self.grid
        
    def winning_player(self):
        return self.previous_player()
        
    def successors(self):
        pnum = self.current_player
        for pos in self.enumerate_coords_with_value(pnum):
            in_zone = self.zone_for_position(pos)==pnum
            nfunc = GridUtils.enumerate_ortho_neighbors if in_zone else GridUtils.enumerate_neighbors
            for npos,nval in nfunc(self.grid, pos):
                if nval!=pnum:
                    nextgrid = self.grid.copy()
                    nextgrid[pos]  = 0
                    nextgrid[npos] = pnum
                    yield ((pos,npos), self.next_state_with_grid(nextgrid))
                    
    def animations_for_move(self, move, newstate):
        import PGPieceAnimation
        return PGPieceAnimation.AnimationGroup(PGPieceAnimation.move_with_capture_animations(move[0], move[1]))
        
#TODO: AI needs to look at distance to goal position
from GenericAI import GenericAI

class ZoneshAI(GenericAI):
    def game_utility(self, state):
        return self.player_utility(state, state.current_player) - self.player_utility(state, state.next_player())
        
    def player_utility(self, state, pnum):
        # utility is number of pieces, reduced slightly for min distance to goal
        def dist(p):
            return max(abs(c) for c in p)
            
        goal = state.extra_state['goal_positions'][pnum-1]
        distances = [dist(GridUtils.subtract_tuples(p, goal)) for p in state.enumerate_coords_with_value(pnum)]
        return 100*len(distances) - min(distances)
        